# The MIT License (MIT)
# Copyright (c) 2021 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal methods for executing functions in the Azure Cosmos database service.
"""
import asyncio  # pylint: disable=do-not-import-asyncio
import json
import time
import logging

from azure.core.exceptions import (AzureError, ClientAuthenticationError, ServiceRequestError,
                                   ServiceResponseError)
from azure.core.pipeline.policies import AsyncRetryPolicy

from .. import _default_retry_policy, _health_check_retry_policy
from .. import _endpoint_discovery_retry_policy
from .. import _gone_retry_policy
from .. import _resource_throttle_retry_policy
from .. import _service_response_retry_policy, _service_request_retry_policy
from .. import _session_retry_policy
from .. import _timeout_failover_retry_policy
from .. import exceptions
from .._constants import _Constants
from .._container_recreate_retry_policy import ContainerRecreateRetryPolicy
from .._retry_utility import (_configure_timeout, _has_read_retryable_headers,
                              _handle_service_response_retries, _handle_service_request_retries,
                              _has_database_account_header)
from ..exceptions import CosmosHttpResponseError
from ..http_constants import HttpHeaders, StatusCodes, SubStatusCodes
from .._cosmos_http_logging_policy import _log_diagnostics_error


# pylint: disable=protected-access, disable=too-many-lines, disable=too-many-statements, disable=too-many-branches

# args [0] is the request object
# args [1] is the connection policy
# args [2] is the pipeline client
# args [3] is the http request
async def ExecuteAsync(client, global_endpoint_manager, function, *args, **kwargs): # pylint: disable=too-many-locals
    """Executes the function with passed parameters applying all retry policies

    :param object client:
        Document client instance
    :param object global_endpoint_manager:
        Instance of _GlobalEndpointManager class
    :param function function:
        Function to be called wrapped with retries
    :param list args:
    :returns: the result of running the passed in function as a (result, headers) tuple
    :rtype: tuple of (dict, dict)
    """
    timeout = kwargs.get('timeout')
    operation_start_time = kwargs.get(_Constants.OperationStartTime, time.time())

    # Track the last error for chaining
    last_error = None

    pk_range_wrapper = None
    if args and global_endpoint_manager.is_circuit_breaker_applicable(args[0]):
        pk_range_wrapper = await global_endpoint_manager.create_pk_range_wrapper(args[0])
    # instantiate all retry policies here to be applied for each request execution
    endpointDiscovery_retry_policy = _endpoint_discovery_retry_policy.EndpointDiscoveryRetryPolicy(
        client.connection_policy, global_endpoint_manager, *args
    )
    health_check_retry_policy = _health_check_retry_policy.HealthCheckRetryPolicy(
        client.connection_policy,
        *args
    )
    resourceThrottle_retry_policy = _resource_throttle_retry_policy.ResourceThrottleRetryPolicy(
        client.connection_policy.RetryOptions.MaxRetryAttemptCount,
        client.connection_policy.RetryOptions.FixedRetryIntervalInMilliseconds,
        client.connection_policy.RetryOptions.MaxWaitTimeInSeconds,
    )
    defaultRetry_policy = _default_retry_policy.DefaultRetryPolicy(*args)

    sessionRetry_policy = _session_retry_policy._SessionRetryPolicy(
        client.connection_policy.EnableEndpointDiscovery, global_endpoint_manager, pk_range_wrapper, *args
    )
    partition_key_range_gone_retry_policy = _gone_retry_policy.PartitionKeyRangeGoneRetryPolicy(client, *args)
    timeout_failover_retry_policy = _timeout_failover_retry_policy._TimeoutFailoverRetryPolicy(
        client.connection_policy, global_endpoint_manager, pk_range_wrapper, *args
    )
    service_response_retry_policy = _service_response_retry_policy.ServiceResponseRetryPolicy(
        client.connection_policy, global_endpoint_manager, pk_range_wrapper, *args,
    )
    service_request_retry_policy = _service_request_retry_policy.ServiceRequestRetryPolicy(
        client.connection_policy, global_endpoint_manager, pk_range_wrapper, *args,
    )
    # Get Logger
    logger = kwargs.get("logger", logging.getLogger("azure.cosmos._retry_utility_async"))
    # HttpRequest we would need to modify for Container Recreate Retry Policy
    request = None
    if args and len(args) > 3:
        # Reference HttpRequest instance in args
        request = args[3]
        container_recreate_retry_policy = ContainerRecreateRetryPolicy(
            client, client._container_properties_cache, request, *args)
    else:
        container_recreate_retry_policy = ContainerRecreateRetryPolicy(
            client, client._container_properties_cache, None, *args)

    while True:
        start_time = time.time()
        # Check timeout before executing function
        if timeout:
            elapsed = time.time() - operation_start_time
            if elapsed >= timeout:
                raise exceptions.CosmosClientTimeoutError(error=last_error)
        try:
            if args:
                result = await ExecuteFunctionAsync(function, global_endpoint_manager, *args, **kwargs)
                await global_endpoint_manager.record_success(args[0])
            else:
                result = await ExecuteFunctionAsync(function, *args, **kwargs)
                # Check timeout after successful execution
                if timeout:
                    elapsed = time.time() - operation_start_time
                    if elapsed >= timeout:
                        raise exceptions.CosmosClientTimeoutError(error=last_error)
            if not client.last_response_headers:
                client.last_response_headers = {}

            # setting the throttle related response headers before returning the result
            client.last_response_headers[
                HttpHeaders.ThrottleRetryCount
            ] = resourceThrottle_retry_policy.current_retry_attempt_count
            client.last_response_headers[
                HttpHeaders.ThrottleRetryWaitTimeInMs
            ] = resourceThrottle_retry_policy.cumulative_wait_time_in_milliseconds
            # TODO: It is better to raise Exceptions manually in the method related to the request,
            #  a rework of retry would be needed to be able to retry exceptions raised that way.
            #  for now raising a manual exception here should allow it to be retried.
            # If container does not have throughput, results will return empty list.
            # We manually raise a 404. We raise it here, so we can handle it in retry utilities.
            if result and isinstance(result[0], dict) and 'Offers' in result[0] and not result[0]['Offers'] \
                    and request.method == 'POST':
                # Grab the link used for getting throughput properties to add to message.
                link = json.loads(request.body)["parameters"][0]["value"]
                response = exceptions._InternalCosmosException(status_code=StatusCodes.NOT_FOUND,
                                                               headers={HttpHeaders.SubStatus:
                                                                     SubStatusCodes.THROUGHPUT_OFFER_NOT_FOUND})
                e_offer = exceptions.CosmosResourceNotFoundError(
                    status_code=StatusCodes.NOT_FOUND,
                    message="Could not find ThroughputProperties for container " + link,
                    response=response)

                response_headers = result[1] if len(result) > 1 else {}
                logger_attributes = {
                    "duration": time.time() - start_time,
                    "verb": request.method,
                    "status_code": e_offer.status_code,
                    "sub_status_code": e_offer.sub_status,
                }
                _log_diagnostics_error(client._enable_diagnostics_logging, request, response_headers, e_offer,
                                       logger_attributes, global_endpoint_manager, logger=logger)
                raise e_offer

            return result
        except exceptions.CosmosHttpResponseError as e:
            last_error = e
            if request:
                # update session token for relevant operations
                client._UpdateSessionIfRequired(request.headers, {}, e.headers)
            if request and _has_database_account_header(request.headers):
                retry_policy = health_check_retry_policy
            elif e.status_code == StatusCodes.FORBIDDEN and e.sub_status in \
                    [SubStatusCodes.DATABASE_ACCOUNT_NOT_FOUND, SubStatusCodes.WRITE_FORBIDDEN]:
                retry_policy = endpointDiscovery_retry_policy
            elif e.status_code == StatusCodes.TOO_MANY_REQUESTS:
                retry_policy = resourceThrottle_retry_policy
            elif (
                e.status_code == StatusCodes.NOT_FOUND
                and e.sub_status
                and e.sub_status == SubStatusCodes.READ_SESSION_NOTAVAILABLE
            ):
                retry_policy = sessionRetry_policy
            elif exceptions._partition_range_is_gone(e):
                retry_policy = partition_key_range_gone_retry_policy
            elif exceptions._container_recreate_exception(e):
                retry_policy = container_recreate_retry_policy
                # Before we retry if retry policy is container recreate, we need refresh the cache of the
                # container properties and pass in the new RID in the headers.
                await client._refresh_container_properties_cache(retry_policy.container_link)
                if e.sub_status != SubStatusCodes.COLLECTION_RID_MISMATCH and retry_policy.check_if_rid_different(
                        retry_policy.container_link, client._container_properties_cache, retry_policy.container_rid):
                    retry_policy.refresh_container_properties_cache = False
                else:
                    cached_container = client._container_properties_cache[retry_policy.container_link]
                    # If partition key value was previously extracted from the document definition
                    # reattempt to extract partition key with updated partition key definition
                    if retry_policy.should_extract_partition_key(cached_container):
                        new_partition_key = await retry_policy._extract_partition_key_async(
                            client, container_cache=cached_container, body=request.body
                        )
                        request.headers[HttpHeaders.PartitionKey] = new_partition_key
                    # If getting throughput, we have to replace the container link received from stale cache
                    # with refreshed cache
                    if retry_policy.should_update_throughput_link(request.body, cached_container):
                        new_body = retry_policy._update_throughput_link(request.body)
                        request.body = new_body

                    retry_policy.container_rid = cached_container["_rid"]
                    request.headers[retry_policy._intended_headers] = retry_policy.container_rid
            elif e.status_code == StatusCodes.REQUEST_TIMEOUT or e.status_code >= StatusCodes.INTERNAL_SERVER_ERROR:
                # record the failure for circuit breaker tracking
                if args:
                    await global_endpoint_manager.record_failure(args[0])
                retry_policy = timeout_failover_retry_policy
            else:
                retry_policy = defaultRetry_policy

            # If none of the retry policies applies or there is no retry needed, set the
            # throttle related response headers and re-throw the exception back arg[0]
            # is the request. It needs to be modified for write forbidden exception
            if not retry_policy.ShouldRetry(e):
                if not client.last_response_headers:
                    client.last_response_headers = {}
                client.last_response_headers[
                    HttpHeaders.ThrottleRetryCount
                ] = resourceThrottle_retry_policy.current_retry_attempt_count
                client.last_response_headers[
                    HttpHeaders.ThrottleRetryWaitTimeInMs
                ] = resourceThrottle_retry_policy.cumulative_wait_time_in_milliseconds
                if args and args[0].should_clear_session_token_on_session_read_failure and client.session:
                    client.session.clear_session_token(client.last_response_headers)
                raise

            # Check timeout only before retrying
            if timeout:
                elapsed = time.time() - operation_start_time
                if elapsed >= timeout:
                    raise exceptions.CosmosClientTimeoutError(error=last_error)
            # Wait for retry_after_in_milliseconds time before the next retry
            await asyncio.sleep(retry_policy.retry_after_in_milliseconds / 1000.0)

        except ServiceRequestError as e:
            if request and _has_database_account_header(request.headers):
                if not health_check_retry_policy.ShouldRetry(e):
                    raise e
            else:
                _handle_service_request_retries(client, service_request_retry_policy, e, *args)

        except ServiceResponseError as e:
            if request and _has_database_account_header(request.headers):
                if not health_check_retry_policy.ShouldRetry(e):
                    raise e
            else:
                try:
                    # pylint: disable=networking-import-outside-azure-core-transport
                    from aiohttp.client_exceptions import (
                        ClientConnectionError)
                    if isinstance(e.inner_exception, ClientConnectionError):
                        _handle_service_request_retries(client, service_request_retry_policy, e, *args)
                    else:
                        if args:
                            await global_endpoint_manager.record_failure(args[0])
                        _handle_service_response_retries(request, client, service_response_retry_policy, e, *args)
                # in case customer is not using aiohttp
                except ImportError:
                    if args:
                        await global_endpoint_manager.record_failure(args[0])
                    _handle_service_response_retries(request, client, service_response_retry_policy, e, *args)


async def ExecuteFunctionAsync(function, *args, **kwargs):
    """Stub method so that it can be used for mocking purposes as well.
    :param Callable function: the function to execute.
    :param list args: the explicit arguments for the function.
    :returns: the result of executing the function with the passed in arguments
    :rtype: tuple(dict, dict)
    """
    return await function(*args, **kwargs)


class _ConnectionRetryPolicy(AsyncRetryPolicy):

    def __init__(self, **kwargs):
        clean_kwargs = {k: v for k, v in kwargs.items() if v is not None}
        super(_ConnectionRetryPolicy, self).__init__(**clean_kwargs)

    async def send(self, request):
        """Sends the PipelineRequest object to the next policy. Uses retry settings if necessary.
        Also enforces an absolute client-side timeout that spans multiple retry attempts.

        :param request: The PipelineRequest object
        :type request: ~azure.core.pipeline.PipelineRequest
        :return: Returns the PipelineResponse or raises error if maximum retries exceeded.
        :rtype: ~azure.core.pipeline.PipelineResponse
        :raises ~azure.core.exceptions.AzureError: Maximum retries exceeded.
        :raises ~azure.cosmos.exceptions.CosmosClientTimeoutError: Specified timeout exceeded.
        :raises ~azure.core.exceptions.ClientAuthenticationError: Authentication failed.
        """
        absolute_timeout = request.context.options.pop('timeout', None)
        per_request_timeout = request.context.options.pop('connection_timeout', 0)
        request_params = request.context.options.pop('request_params', None)
        global_endpoint_manager = request.context.options.pop('global_endpoint_manager', None)
        retry_error = None
        retry_active = True
        response = None
        retry_settings = self.configure_retries(request.context.options)
        while retry_active:
            start_time = time.time()
            try:
                _configure_timeout(request, absolute_timeout, per_request_timeout)
                response = await self.next.send(request)
                break
            except ClientAuthenticationError:  # pylint:disable=try-except-raise
                # the authentication policy failed such that the client's request can't
                # succeed--we'll never have a response to it, so propagate the exception
                raise
            except exceptions.CosmosClientTimeoutError as timeout_error:
                timeout_error.inner_exception = retry_error
                timeout_error.response = response
                timeout_error.history = retry_settings['history']
                raise
            except ServiceRequestError as err:
                retry_error = err
                # the request ran into a socket timeout or failed to establish a new connection
                # since request wasn't sent, raise exception immediately to be dealt with in client retry policies
                if (not _has_database_account_header(request.http_request.headers)
                        and not request_params.healthy_tentative_location):
                    if retry_settings['connect'] > 0:
                        retry_active = self.increment(retry_settings, response=request, error=err)
                        if retry_active:
                            await self.sleep(retry_settings, request.context.transport)
                            continue
                raise err
            except ServiceResponseError as err:
                retry_error = err
                if (_has_database_account_header(request.http_request.headers) or
                        request_params.healthy_tentative_location):
                    raise err
                # Since this is ClientConnectionError, it is safe to be retried on both read and write requests
                try:
                    # pylint: disable=networking-import-outside-azure-core-transport
                    from aiohttp.client_exceptions import (
                        ClientConnectionError)
                    if (isinstance(err.inner_exception, ClientConnectionError)
                            or _has_read_retryable_headers(request.http_request.headers)):
                        # This logic is based on the _retry.py file from azure-core
                        if retry_settings['read'] > 0:
                            # record the failure for circuit breaker tracking for retries in connection retry policy
                            # retries in the execute function will mark those failures
                            await global_endpoint_manager.record_failure(request_params)
                            retry_active = self.increment(retry_settings, response=request, error=err)
                            if retry_active:
                                await self.sleep(retry_settings, request.context.transport)
                                continue

                except ImportError:
                    raise err # pylint: disable=raise-missing-from
                raise err
            except CosmosHttpResponseError as err:
                raise err
            except AzureError as err:
                retry_error = err
                if (_has_database_account_header(request.http_request.headers) or
                        request_params.healthy_tentative_location):
                    raise err
                if _has_read_retryable_headers(request.http_request.headers) and retry_settings['read'] > 0:
                    retry_active = self.increment(retry_settings, response=request, error=err)
                    if retry_active:
                        await self.sleep(retry_settings, request.context.transport)
                        continue
                raise err
            finally:
                end_time = time.time()
                if absolute_timeout:
                    absolute_timeout -= (end_time - start_time)

        self.update_context(response.context, retry_settings)
        return response



# -*- coding: utf-8 -*-
# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

"""End-to-end test.
"""
import copy
import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.exceptions as exceptions
import test_config
from azure.cosmos.partition_key import PartitionKey


@pytest.mark.cosmosLong
class TestContainerPropertiesCache(unittest.TestCase):
    """Python CRUD Tests.
        """

    configs = test_config.TestConfig
    host = configs.host
    masterKey = configs.masterKey
    connectionPolicy = configs.connectionPolicy
    client: cosmos_client.CosmosClient = None


    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")
        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.databaseForTest = cls.client.create_database_if_not_exists(cls.configs.TEST_DATABASE_ID)

    def test_container_properties_cache(self):
        client = self.client
        database_name = self.configs.TEST_DATABASE_ID
        created_db = self.databaseForTest
        container_name = "Container Properties Cache Test " + str(uuid.uuid4())
        container_pk = "PK"
        # Create The Container
        try:
            client.get_database_client(database_name).create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # Delete The cache as this is meant to test calling operations on a preexisting container
        # and not a freshly made one. It's a private attribute so use mangled name.
        client.client_connection._CosmosClientConnection__container_properties_cache = {}
        # We will hot path operations to verify cache persists
        # This will extract partition key from the item body, which will need partition key definition from
        # container properties. We test to check the cache is empty since we just created the container
        self.assertTrue(client.client_connection._container_properties_cache == {})
        client.get_database_client(database_name).get_container_client(container_name).create_item(
            body={'id': 'item1', container_pk: 'value'})
        # Since the cache was empty, it should have called a container read to get properties. So now Cache should
        # be populated and available even when we don't have a container instance
        self.assertTrue(client.client_connection._container_properties_cache != {})
        # We can test if the cache properties are correct by comparing them to a fresh read.
        # First lets save the old cache values
        cached_properties = created_db.get_container_client(container_name)._get_properties()
        # Get the container dictionary out of a fresh container read
        fresh_container_read = created_db.get_container_client(container_name).read()
        # Now we can compare the RID and Partition Key Definition
        self.assertEqual(cached_properties.get("_rid"), fresh_container_read.get("_rid"))
        self.assertEqual(cached_properties.get("partitionKey"), fresh_container_read.get("partitionKey"))
        created_db.delete_container(container_name)

    def test_container_recreate_create_upsert_replace_item(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # This Simulates a container recreate. We save the old cache and then create
        # a new container with different partition key definition
        # for these three operations we extract the partition key value from the document
        # with a stale cache we end up extracting the wrong one so these will retry extracting
        # the partition key after refreshing the cache. Test to make sure a container recreate doesn't affect it.
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # let's test create by replacing current container properties cache with old one
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.create_item(body={'id': 'item1', container2_pk: 'val'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        # Do the same thing for upsert operation
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.upsert_item(dict(id="item2", partkey="value"))
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        # Finally test replace item
        item = created_container.read_item("item1", partition_key="val")
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.replace_item(item=item, body={'id': 'item1', container2_pk: 'val'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Now a negative test. We will try creating an item as if we were attempting to create it with the old container
        # This should result in an error of 400 as the container has been recreated.
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.create_item(body={'id': 'item3', container_pk: 'val'})
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 400)
        created_db.delete_container(container_name)

    def test_container_recreate_create_upsert_replace_item_sub_partitioning(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = ["/country", "/state"]
        container2_pk = ["/county", "/city"]
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # This Simulates a container recreate. We save the old cache and then create
        # a new container with different partition key definition
        # for these three operations we extract the partition key value from the document
        # with a stale cache we end up extracting the wrong one so these will retry extracting
        # the partition key after refreshing the cache. Test to make sure a container recreate doesn't affect it.
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # let's test create by replacing current container properties cache with old one
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.create_item(body={'id': 'item1', 'county': 'ventura', 'city': 'oxnard'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        # Do the same thing for upsert operation
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.upsert_item(dict(id="item2", county='Santa Barbara', city='Santa Barbara'))
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        # Finally test replace item
        item = created_container.read_item("item1", partition_key=['ventura', 'oxnard'])
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.replace_item(item=item, body={'id': 'item1', 'county': 'ventura', 'city': 'oxnard'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Now a negative test. We will try creating an item as if we were attempting to create it with the old container
        # This should result in an error of 400 as the container has been recreated.
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            created_container.create_item(body={'id': 'item3', 'country': 'USA', 'state': 'CA'})
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 400)
        created_db.delete_container(container_name)

    def test_offer_throughput_container_recreate(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk), offer_throughput=600)
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # This Simulates a container recreate. We save the old cache and then create
        # a new container with different container properties
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container2_pk), offer_throughput=800)
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # let's test offer throughput by replacing current container properties cache with old one
        # Check to make sure it equals the new created container's throughput.
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            offer = created_container.read_offer()
            self.assertEqual(offer.offer_throughput, 800)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Now check for replace throughput
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        new_throughput = 900
        try:
            new_offer = created_container.replace_throughput(new_throughput)
            self.assertEqual(new_offer.offer_throughput, new_throughput)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        created_db.delete_container(container_name)

    def test_offer_throughput_container_recreate_sub_partition(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = ["/country", "/state"]
        container2_pk = ["/county", "/city"]
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container_pk), offer_throughput=600)
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # This Simulates a container recreate. We save the old cache and then create
        # a new container with different container properties
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container2_pk), offer_throughput=800)
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # let's test offer throughput by replacing current container properties cache with old one
        # Check to make sure it equals the new created container's throughput.
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            offer = created_container.read_offer()
            self.assertEqual(offer.offer_throughput, 800)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Now check for replace throughput
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        new_throughput = 900
        try:
            new_offer = created_container.replace_throughput(new_throughput)
            self.assertEqual(new_offer.offer_throughput, new_throughput)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        created_db.delete_container(container_name)

    def test_container_recreate_read_item(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # create an item to read
        try:
            item_to_read = created_container.create_item(body={'id': 'item1', container_pk: 'val'})
            item_to_read2 = created_container.create_item(body={'id': 'item2', container_pk: 'OtherValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            created_container.create_item(body={'id': 'item1', container2_pk: 'val'})
            created_container.create_item(body={'id': 'item2', container2_pk: 'DifferentValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        # Will fail because original container is deleted
        try:
            created_container.read_item(item=item_to_read, partition_key="val")
            self.fail("Read should not succeed as item no longer exists.")
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 404)
        # Will succeed because item id and partition key value are the same as an existing one in the new container
        # But read items are not the same as their RID will be different
        read_item = created_container.read_item(item="item1", partition_key="val")
        self.assertNotEqual(read_item, item_to_read)
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            # Cache will refresh but will error out since item no longer exists
            created_container.read_item(item=item_to_read2, partition_key="OtherValue")
            self.fail("Read should not succeed as item no longer exists.")
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 404)
        created_db.delete_container(container_name)

    def test_container_recreate_read_item_sub_partition(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = ["/country", "/state"]
        container2_pk = ["/county", "/city"]
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            item_to_read = created_container.create_item(body={'id': 'item1', 'country': 'USA', 'state': 'CA'})
            item_to_read2 = created_container.create_item(body={'id': 'item2',
                                                                'country': 'MEX',
                                                                'state': 'Michoacán'})  # cspell:disable-line
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        # Recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            created_container.create_item(body={'id': 'item1', 'county': 'ventura', 'city': 'oxnard'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            # Cache will refresh but will error out since item no longer exists
            created_container.read_item(item=item_to_read, partition_key=['USA', 'CA'])
            self.fail("Read should not succeed as item no longer exists.")
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 404)
        created_db.delete_container(container_name)

    def test_container_recreate_delete_item(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # create items
        try:
            item_to_delete = created_container.create_item(body={'id': 'item1', container_pk: 'val'})
            created_container.create_item(body={'id': 'item2', container_pk: 'OtherValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            created_container.create_item(body={'id': 'item1', container2_pk: 'val'})
            created_container.create_item(body={'id': 'item2', container2_pk: 'DifferentValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        # Should fail as the item no longer exists
        try:
            created_container.delete_item(item_to_delete, partition_key='val')
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 404)
        created_db.delete_container(container_name)

    def test_container_recreate_delete_item_sub_partition(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = ["/country", "/state"]
        container2_pk = ["/county", "/city"]
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # create items
        try:
            item_to_del = created_container.create_item(body={'id': 'item1', 'country': 'USA', 'state': 'CA'})
            created_container.create_item(body={'id': 'item2',
                                                'country': 'MEX', 'state': 'Michoacán'})  # cspell:disable-line
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path=container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            created_container.create_item(body={'id': 'item1', 'county': 'ventura', 'city': 'oxnard'})
            created_container.create_item(body={'id': 'item2',
                                                               'county': 'santa barbara', 'city': 'santa barbara'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        # Should fail as the item no longer exists
        try:
            created_container.delete_item(item_to_del, partition_key=['USA', 'CA'])
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 404)
        created_db.delete_container(container_name)

    def test_container_recreate_query(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"
        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        # create an item to read
        try:
            item_to_read = created_container.create_item(body={'id': 'item1', container_pk: 'val'})
            item_to_read2 = created_container.create_item(body={'id': 'item2', container_pk: 'OtherValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name, partition_key=PartitionKey(
                path="/" + container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")
        try:
            created_container.create_item(body={'id': 'item1', container2_pk: 'val'})
            created_container.create_item(body={'id': 'item2', container2_pk: 'DifferentValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        # Will succeed because item id and partition key value are the same as an existing one in the new container
        # But queried items are not the same as their RID will be different
        try:
            query = "SELECT * FROM c WHERE c.id = @id"
            parameters = [{"name": "@id", "value": item_to_read['id']}]
            query_result = list(created_container.query_items(query=query,
                                                              parameters=parameters,
                                                              enable_cross_partition_query=True))
            self.assertEqual(1, len(query_result))
        except exceptions.CosmosHttpResponseError as e:
            self.fail("Query should still succeed as item has same id despite the RID not matching.")
        # Will succeed because item id and partition key value are the same as an existing one in the new container
        # But queried items are not the same as their RID will be different
        query = "SELECT * FROM c WHERE c.id = @id"
        parameters = [{"name": "@id", "value": 'item1'}]
        query_result = list(
            created_container.query_items(query=query, parameters=parameters, enable_cross_partition_query=True))
        self.assertNotEqual(query_result[0], item_to_read)
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            # Trying to query for item from old container will return empty query
            query = "SELECT * FROM c WHERE c.partKey = @partKey"
            parameters = [{"name": "@partKey", "value": item_to_read2['PK']}]
            query_result = list(created_container.query_items(query=query,
                                                              parameters=parameters,
                                                              enable_cross_partition_query=True))
            self.assertEqual(0, len(query_result))
        except exceptions.CosmosHttpResponseError as e:
            self.fail("Query should still succeed if container is recreated.")
        created_db.delete_container(container_name)

    def test_container_recreate_transactional_batch(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"
        container2_pk = "partkey"

        # Create The Container
        try:
            created_container = created_db.create_container(id=container_name,
                                                            partition_key=PartitionKey(path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        try:
            batch_operations = []
            for i in range(100):
                batch_operations.append(("create", ({"id": "item" + str(i), container_pk: "Microsoft"},)))

            batch_response = created_container.execute_item_batch(batch_operations=batch_operations,
                                                                  partition_key="Microsoft")
            self.assertEqual(len(batch_response), 100)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Simulate a container recreate by saving the old cache and creating a new container with a different partition key definition
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name,
                                                            partition_key=PartitionKey(path="/" + container2_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container Should not Already Exist.")

        # Test transactional batch operation by replacing current container properties cache with old one
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            batch_operations = []
            for i in range(100):
                batch_operations.append(("create", ({"id": "item" + str(i), container2_pk: "Microsoft"},)))

            batch_response = created_container.execute_item_batch(batch_operations=batch_operations,
                                                                  partition_key="Microsoft")
            self.assertEqual(len(batch_response), 100)
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Negative test: Attempt to create an item with the old container properties cache, expecting a 400 error
        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)
        try:
            batch_operations = [("create", ({"id": "item" + str(101), container2_pk: "Microsoft"},))]
            batch_response = created_container.execute_item_batch(batch_operations=batch_operations,
                                                                  partition_key="Microsoft")
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, 400)

        created_db.delete_container(container_name)

    def test_container_recreate_change_feed(self):
        client = self.client
        created_db = self.databaseForTest
        container_name = str(uuid.uuid4())
        container_pk = "PK"

        # Create the container
        try:
            created_container = created_db.create_container(id=container_name,
                                                            partition_key=PartitionKey(path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container should not already exist.")

        # Create initial items
        try:
            item_to_read = created_container.create_item(body={'id': 'item1', container_pk: 'val'})
            item_to_read2 = created_container.create_item(body={'id': 'item2', container_pk: 'OtherValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        # Save old container cache and recreate container
        old_cache = copy.deepcopy(client.client_connection._CosmosClientConnection__container_properties_cache)
        created_db.delete_container(created_container)
        try:
            created_container = created_db.create_container(id=container_name,
                                                            partition_key=PartitionKey(path="/" + container_pk))
        except exceptions.CosmosResourceExistsError:
            self.fail("Container should not already exist.")

        # Create new items in the recreated container
        try:
            new_item1 = created_container.create_item(body={'id': 'item3', container_pk: 'newVal'})
            new_item2 = created_container.create_item(body={'id': 'item4', container_pk: 'newOtherValue'})
        except exceptions.CosmosHttpResponseError as e:
            self.fail("{}".format(e.http_error_message))

        client.client_connection._CosmosClientConnection__container_properties_cache = copy.deepcopy(old_cache)

        # Query change feed for the new items
        change_feed = list(created_container.query_items_change_feed(start_time='Beginning'))
        self.assertEqual(len(change_feed), 2)

        # Verify that the change feed contains the new items
        self.assertTrue(any(item['id'] == 'item3' and item[container_pk] == 'newVal' for item in change_feed))
        self.assertTrue(any(item['id'] == 'item4' and item[container_pk] == 'newOtherValue' for item in change_feed))

        # Verify that the change feed does not contain the old items
        self.assertFalse(any(item['id'] == 'item1' and item[container_pk] == 'val' for item in change_feed))
        self.assertFalse(any(item['id'] == 'item2' and item[container_pk] == 'OtherValue' for item in change_feed))

        created_db.delete_container(container_name)


if __name__ == '__main__':
    try:
        unittest.main()
    except SystemExit as inst:
        if inst.args[0] is True:  # raised by sys.exit(True) when tests failed
            raise



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.cosmos.aio import CosmosClient
from azure.cosmos.partition_key import PartitionKey
import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates how to configure custom user agents for async Azure Cosmos DB clients
#
# User agents help identify which application or process is making requests to Cosmos DB.
# Custom user agents are particularly useful for:
#   - Identifying different applications using the same Cosmos DB account
#   - Distinguishing between different versions of your application
#   - Tracking requests from specific microservices or processes
#   - Debugging and monitoring application behavior in diagnostics
#   - Analyzing usage patterns across different client applications
#
# The user agent string appears in Cosmos DB diagnostics, logs, and monitoring tools,
# making it easier to pinpoint which client instance is making specific requests.
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']


async def create_client_with_default_user_agent():
    """
    Create an async Cosmos DB client with the default user agent.
    
    The default user agent includes information about the SDK version and Python version.
    Example: "azsdk-python-cosmos/4.5.0 Python/3.11.0 (Windows-10-10.0.22621-SP0)"
    """
    print('\n1. Creating async client with default user agent')
    
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        print('Async client created with default user agent')
        print('The default user agent includes SDK version, Python version, and platform information')
        return client


async def create_client_with_custom_user_agent():
    """
    Create an async Cosmos DB client with a custom user agent suffix.
    
    The user_agent parameter appends a custom string to the default user agent,
    allowing you to identify your specific application or service.
    """
    print('\n2. Creating async client with custom user agent')
    
    # Add a custom suffix to identify your application
    custom_user_agent = "MyAsyncApplication/1.0.0"
    
    async with CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent=custom_user_agent
    ) as client:
        print(f'Async client created with custom user agent: {custom_user_agent}')
        print('This will appear as: [default-user-agent] {}'.format(custom_user_agent))
        return client


async def create_client_with_detailed_user_agent():
    """
    Create an async client with a detailed custom user agent including multiple identifiers.
    
    You can include multiple pieces of information to make diagnostics more useful:
    - Application name and version
    - Service or microservice name
    - Environment (dev, staging, production)
    - Instance or process identifier
    """
    print('\n3. Creating async client with detailed custom user agent')
    
    # Include detailed information about your application
    app_name = "AsyncOrderProcessingService"
    app_version = "2.3.1"
    environment = "production"
    instance_id = "async-instance-42"
    
    detailed_user_agent = f"{app_name}/{app_version} env:{environment} {instance_id}"
    
    async with CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent=detailed_user_agent
    ) as client:
        print(f'Async client created with detailed user agent: {detailed_user_agent}')
        print('This helps identify specific instances in diagnostics and logs')
        return client


async def create_multiple_clients_with_different_user_agents():
    """
    Demonstrate creating multiple async clients with different user agents.
    
    This is useful when you have multiple services or components in your application
    that need to be tracked separately in diagnostics.
    """
    print('\n4. Creating multiple async clients with different user agents')
    
    # Note: In production, you would manage these clients' lifecycles appropriately
    # Here we demonstrate the creation pattern
    
    print('Creating async client for data ingestion: AsyncDataIngestionService/1.0.0')
    print('Creating async client for queries: AsyncQueryService/1.0.0')
    print('Creating async client for analytics: AsyncAnalyticsService/1.0.0')
    
    print('\nEach client can be created with its own user agent to track requests separately')


async def demonstrate_user_agent_in_operations():
    """
    Perform async operations with a custom user agent and show how it helps with diagnostics.
    
    The user agent will appear in:
    - Azure Monitor logs
    - Cosmos DB diagnostic logs
    - Request headers sent to the service
    - Performance and usage analytics
    """
    print('\n5. Demonstrating user agent in async operations')
    
    custom_suffix = "AsyncDemoClient/1.0.0"
    
    async with CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent=custom_suffix
    ) as client:
        try:
            # Create database
            db = await client.create_database_if_not_exists(id=DATABASE_ID)
            print(f'Database created/accessed with user agent: {custom_suffix}')
            
            # Create container
            container = await db.create_container_if_not_exists(
                id=CONTAINER_ID,
                partition_key=PartitionKey(path='/id')
            )
            print(f'Container created/accessed with user agent: {custom_suffix}')
            
            # Create or upsert an item
            item = {
                'id': 'async_sample_item_1',
                'name': 'Async Sample Item',
                'description': 'Created with custom user agent in async operation'
            }
            await container.upsert_item(body=item)
            print(f'Item created/updated with user agent: {custom_suffix}')
            
            # Query items
            query_iterable = container.query_items(
                query="SELECT * FROM c WHERE c.id = @id",
                parameters=[{"name": "@id", "value": "async_sample_item_1"}]
            )
            items = [item async for item in query_iterable]
            print(f'Query executed with user agent: {custom_suffix}')
            print(f'Found {len(items)} item(s)')
            
            print('\nAll these async operations will show your custom user agent in diagnostics')
            
            # Cleanup: Delete the item
            await container.delete_item(item='async_sample_item_1', partition_key='async_sample_item_1')
            print(f'\nCleanup: Deleted sample item')
            
        except Exception as e:
            print(f'Error during async operations: {e}')


async def demonstrate_concurrent_operations_with_user_agents():
    """
    Show how different user agents help track concurrent async operations.
    """
    print('\n6. Demonstrating concurrent operations with different user agents')
    
    async def operation_with_user_agent(user_agent_suffix, operation_name):
        """Helper function to perform operation with specific user agent."""
        async with CosmosClient(
            HOST,
            {'masterKey': MASTER_KEY},
            user_agent=user_agent_suffix
        ) as client:
            db = await client.create_database_if_not_exists(id=DATABASE_ID)
            print(f'{operation_name} completed with user agent: {user_agent_suffix}')
    
    # Run multiple operations concurrently with different user agents
    await asyncio.gather(
        operation_with_user_agent("AsyncWorker1/1.0.0", "Worker 1"),
        operation_with_user_agent("AsyncWorker2/1.0.0", "Worker 2"),
        operation_with_user_agent("AsyncWorker3/1.0.0", "Worker 3")
    )
    
    print('\nEach concurrent operation can be tracked separately in diagnostics')


async def run_sample():
    """
    Run all async user agent examples.
    """
    print('=' * 70)
    print('Azure Cosmos DB - Async User Agent Management Sample')
    print('=' * 70)
    
    # Example 1: Default user agent
    await create_client_with_default_user_agent()
    
    # Example 2: Simple custom user agent
    await create_client_with_custom_user_agent()
    
    # Example 3: Detailed user agent
    await create_client_with_detailed_user_agent()
    
    # Example 4: Multiple clients with different user agents
    await create_multiple_clients_with_different_user_agents()
    
    # Example 5: Show user agent in operations
    await demonstrate_user_agent_in_operations()
    
    # Example 6: Concurrent operations
    await demonstrate_concurrent_operations_with_user_agents()
    
    print('\n' + '=' * 70)
    print('Sample completed!')
    print('=' * 70)


if __name__ == '__main__':
    asyncio.run(run_sample())



# The MIT License (MIT)
# Copyright (c) 2018 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the abstraction to resolve target location for geo-replicated
DatabaseAccount with multiple writable and readable locations.
"""
import collections
import logging
from typing import Set, Mapping, OrderedDict
from urllib.parse import urlparse

from . import documents, _base as base
from .http_constants import ResourceType
from .documents import ConnectionPolicy
from ._request_object import RequestObject

# pylint: disable=protected-access

logger = logging.getLogger("azure.cosmos.LocationCache")

class EndpointOperationType(object):
    NoneType = "None"
    ReadType = "Read"
    WriteType = "Write"

class RegionalRoutingContext(object):
    def __init__(self, primary_endpoint: str):
        self.primary_endpoint: str = primary_endpoint

    def set_primary(self, endpoint: str):
        self.primary_endpoint = endpoint

    def get_primary(self):
        return self.primary_endpoint

    def __eq__(self, other):
        return self.primary_endpoint == other.primary_endpoint

    def __str__(self):
        return "Primary: " + self.primary_endpoint

def get_regional_routing_contexts_by_loc(new_locations: list[dict[str, str]]):
    # construct from previous object
    regional_routing_contexts_by_location: OrderedDict[str, RegionalRoutingContext] = collections.OrderedDict()
    parsed_locations = []

    for new_location in new_locations:
        # if name in new_location and same for database account endpoint
        if "name" in new_location and "databaseAccountEndpoint" in new_location:
            if not new_location["name"]:
                # during fail-over the location name is empty
                continue
            try:
                region_uri = new_location["databaseAccountEndpoint"]
                parsed_locations.append(new_location["name"])
                regional_object = RegionalRoutingContext(region_uri)
                regional_routing_contexts_by_location.update({new_location["name"]: regional_object})
            except Exception as e:
                raise e

    # Also store a hash map of endpoints for each location
    locations_by_endpoints = {value.get_primary(): key for key, value in regional_routing_contexts_by_location.items()}

    return regional_routing_contexts_by_location, locations_by_endpoints, parsed_locations

def _get_health_check_endpoints(regional_routing_contexts) -> Set[str]:
    # should use the endpoints in the order returned from gateway and only the ones specified in preferred locations
    preferred_endpoints = {context.get_primary() for context in regional_routing_contexts}
    return preferred_endpoints

def _get_applicable_regional_routing_contexts(regional_routing_contexts: list[RegionalRoutingContext],
                                              location_name_by_endpoint: Mapping[str, str],
                                              fall_back_regional_routing_context: RegionalRoutingContext,
                                              exclude_location_list: list[str],
                                              circuit_breaker_exclude_list: list[str],
                                              resource_type: str) -> list[RegionalRoutingContext]:
    """Filters and reorders regional endpoints based on exclusion lists and health.

    This method separates the initial list of endpoints into two groups: those the user has explicitly excluded
    and those they have not. It then takes the list of non-excluded endpoints and moves any that are currently
    marked as unavailable by the circuit breaker to the end of that list. This ensures healthy endpoints are
    tried before unhealthy ones.

    For special metadata requests (which must succeed), it adds the user-excluded locations back to the very
    end of the list. This allows the SDK to try every possible endpoint as a last resort for critical operations.

    If all available endpoints are filtered out, it adds a default fallback endpoint to the list to ensure
    there is always at least one endpoint to attempt a connection to.

    :param regional_routing_contexts: The initial list of regional contexts to filter.
    :type regional_routing_contexts: list[RegionalRoutingContext]
    :param location_name_by_endpoint: A mapping from endpoint URL to location name.
    :type location_name_by_endpoint: Mapping[str, str]
    :param fall_back_regional_routing_context: The context to use as a fallback if all others are filtered out.
    :type fall_back_regional_routing_context: RegionalRoutingContext
    :param exclude_location_list: A list of location names to exclude, based on user configuration.
    :type exclude_location_list: list[str]
    :param circuit_breaker_exclude_list: A list of location names to temporarily exclude due to circuit breaker logic.
    :type circuit_breaker_exclude_list: list[str]
    :param resource_type: The type of resource for the request, used to determine if it's a metadata request.
    :type resource_type: str
    :return: A filtered and reordered list of regional routing contexts.
    :rtype: list[RegionalRoutingContext]
    """
    # filter endpoints by excluded locations
    applicable_regional_routing_contexts = []
    user_excluded_regional_routing_contexts = []
    for regional_routing_context in regional_routing_contexts:
        if location_name_by_endpoint.get(regional_routing_context.get_primary()) not in exclude_location_list:
            applicable_regional_routing_contexts.append(regional_routing_context)
        else:
            user_excluded_regional_routing_contexts.append(regional_routing_context)

    # Now, filter by circuit breaker exclusions, moving them to the end of the list
    final_applicable_contexts = []
    circuit_breaker_excluded_contexts = []
    for regional_routing_context in applicable_regional_routing_contexts:
        if location_name_by_endpoint.get(regional_routing_context.get_primary()) in circuit_breaker_exclude_list:
            circuit_breaker_excluded_contexts.append(regional_routing_context)
        else:
            final_applicable_contexts.append(regional_routing_context)

    # For metadata requests, add user-excluded locations BEFORE fallback
    if base.IsMasterResource(resource_type):
        final_applicable_contexts.extend(user_excluded_regional_routing_contexts)

    # If no healthy regions, try circuit-breaker excluded ones BEFORE global fallback
    if not final_applicable_contexts and circuit_breaker_excluded_contexts:
        final_applicable_contexts = circuit_breaker_excluded_contexts
    elif not final_applicable_contexts:
        # Only use global fallback if there are no other options
        final_applicable_contexts.append(fall_back_regional_routing_context)

    return final_applicable_contexts

class LocationCache(object):  # pylint: disable=too-many-public-methods,too-many-instance-attributes

    def __init__(
        self,
        default_endpoint: str,
        connection_policy: ConnectionPolicy,
    ):
        self.default_regional_routing_context: RegionalRoutingContext = RegionalRoutingContext(default_endpoint)
        self.effective_preferred_locations: list[str] = []
        self.enable_multiple_writable_locations: bool = False
        self.write_regional_routing_contexts: list[RegionalRoutingContext] = [self.default_regional_routing_context]
        self.read_regional_routing_contexts: list[RegionalRoutingContext] = [self.default_regional_routing_context]
        self.location_unavailability_info_by_endpoint: dict[str, dict[str, Set[EndpointOperationType]]] = {}
        self.last_cache_update_time_stamp: int = 0
        self.account_read_regional_routing_contexts_by_location: dict[str, RegionalRoutingContext] = {} # pylint: disable=name-too-long
        self.account_write_regional_routing_contexts_by_location: dict[str, RegionalRoutingContext] = {} # pylint: disable=name-too-long
        self.account_locations_by_read_endpoints: dict[str, str] = {} # pylint: disable=name-too-long
        self.account_locations_by_write_endpoints: dict[str, str] = {} # pylint: disable=name-too-long
        self.account_write_locations: list[str] = []
        self.account_read_locations: list[str] = []
        self.connection_policy: ConnectionPolicy = connection_policy

    def get_write_regional_routing_contexts(self):
        return self.write_regional_routing_contexts

    def get_read_regional_routing_contexts(self):
        return self.read_regional_routing_contexts

    def get_location_from_endpoint(self, endpoint: str) -> str:
        if endpoint in self.account_locations_by_read_endpoints:
            return self.account_locations_by_read_endpoints[endpoint]
        return self.account_write_locations[0]

    def get_write_regional_routing_context(self):
        return self.get_write_regional_routing_contexts()[0].get_primary()

    def get_read_regional_routing_context(self):
        return self.get_read_regional_routing_contexts()[0].get_primary()

    def mark_endpoint_unavailable_for_read(self, endpoint, refresh_cache, context="Unknown"):
        self.mark_endpoint_unavailable(endpoint, EndpointOperationType.ReadType, refresh_cache, context)

    def mark_endpoint_unavailable_for_write(self, endpoint, refresh_cache, context="Unknown"):
        self.mark_endpoint_unavailable(endpoint, EndpointOperationType.WriteType, refresh_cache, context)

    def perform_on_database_account_read(self, database_account):
        self.update_location_cache(
            database_account._WritableLocations,
            database_account._ReadableLocations,
            database_account._EnableMultipleWritableLocations,
        )

    def get_all_write_endpoints(self) -> Set[str]:
        return {
            context.get_primary()
            for context in self.get_write_regional_routing_contexts()
        }

    def get_ordered_write_locations(self):
        return self.account_write_locations

    def get_ordered_read_locations(self):
        return self.account_read_locations

    def _get_configured_excluded_locations(self, request: RequestObject) -> list[str]:
        # If excluded locations were configured on request, use request level excluded locations.
        excluded_locations = request.excluded_locations
        if excluded_locations is None:
            if self.connection_policy.ExcludedLocations:
                # If excluded locations were only configured on client(connection_policy), use client level
                # make copy of excluded locations to avoid modifying the original list
                excluded_locations = list(self.connection_policy.ExcludedLocations)
            else:
                excluded_locations = []

        return excluded_locations

    def _get_applicable_read_regional_routing_contexts(self, request: RequestObject) -> list[RegionalRoutingContext]:
        # Get configured excluded locations
        excluded_locations = self._get_configured_excluded_locations(request)

        # If excluded locations were configured, return filtered regional endpoints by excluded locations.
        if excluded_locations or request.excluded_locations_circuit_breaker:
            return _get_applicable_regional_routing_contexts(
                self.get_read_regional_routing_contexts(),
                self.account_locations_by_read_endpoints,
                self.get_write_regional_routing_contexts()[0],
                excluded_locations,
                request.excluded_locations_circuit_breaker or [],
                request.resource_type)

        # Else, return all regional endpoints
        return self.get_read_regional_routing_contexts()

    def _get_applicable_write_regional_routing_contexts(self, request: RequestObject) -> list[RegionalRoutingContext]:
        # Get configured excluded locations
        excluded_locations = self._get_configured_excluded_locations(request)

        # If excluded locations were configured, return filtered regional endpoints by excluded locations.
        if excluded_locations or request.excluded_locations_circuit_breaker:
            return _get_applicable_regional_routing_contexts(
                self.get_write_regional_routing_contexts(),
                self.account_locations_by_write_endpoints,
                self.default_regional_routing_context,
                excluded_locations,
                request.excluded_locations_circuit_breaker or [],
                request.resource_type)

        # Else, return all regional endpoints
        return self.get_write_regional_routing_contexts()

    def _resolve_endpoint_without_preferred_locations(self, request, is_write, location_index):
        """Resolves an endpoint when not using preferred locations or for single-write failover.

        This helper method is called when `use_preferred_locations` is False or for write operations on single-write
        accounts. It determines the appropriate endpoint by cycling through available locations while respecting
        user-configured and circuit-breaker-based exclusion lists.

        :param request: The request object for the current operation.
        :type request: azure.cosmos._request_object.RequestObject
        :param is_write: A boolean indicating if the operation is a write.
        :type is_write: bool
        :param location_index: The index used to select an endpoint from the list of available locations.
        :type location_index: int
        :return: The resolved endpoint URL as a string.
        :rtype: str
        """
        ordered_locations = self.account_write_locations if is_write else self.account_read_locations
        all_contexts_by_loc = (self.account_write_regional_routing_contexts_by_location if is_write
                               else self.account_read_regional_routing_contexts_by_location)

        # Safety check: if endpoint discovery is off or location cache isn't populated, fallback.
        if not self.connection_policy.EnableEndpointDiscovery or not ordered_locations:
            return self.default_regional_routing_context.get_primary()

        # For single-write-region accounts, failover between the first two available write regions.
        if is_write and not self.can_use_multiple_write_locations_for_request(request):
            effective_index = min(location_index % 2, len(ordered_locations) - 1)
            write_location = ordered_locations[effective_index]
            if write_location in all_contexts_by_loc:
                return all_contexts_by_loc[write_location].get_primary()
            return self.default_regional_routing_context.get_primary()  # Fallback if location not found

        # For reads or multi-write, filter locations by user and circuit-breaker exclusions.
        excluded_locations = self._get_configured_excluded_locations(request)
        circuit_breaker_excluded_locations = request.excluded_locations_circuit_breaker or []

        applicable_contexts = []
        circuit_breaker_contexts = []
        for loc_name in ordered_locations:
            if loc_name in all_contexts_by_loc:
                context = all_contexts_by_loc[loc_name]
                if loc_name in excluded_locations:
                    continue  # Skip user-excluded locations
                if loc_name in circuit_breaker_excluded_locations:
                    circuit_breaker_contexts.append(context)
                else:
                    applicable_contexts.append(context)

        # Only add circuit breaker excluded locations if no healthy regions exist
        if not applicable_contexts and circuit_breaker_contexts:
            applicable_contexts = circuit_breaker_contexts

        if applicable_contexts:
            effective_index = location_index % len(applicable_contexts)
            return applicable_contexts[effective_index].get_primary()

        # Fallback to the default endpoint if no other endpoint is found.
        return self.default_regional_routing_context.get_primary()

    def resolve_service_endpoint(self, request):
        """Determines the appropriate service endpoint for a given request.

        This method intelligently routes requests by following a specific logic:
        1.  If `request.location_endpoint_to_route` is set, it is used immediately.
        2.  No Preferred Locations or Single-Write Failover: If `use_preferred_locations` is False or it's a
            write operation on a single-write account, it calls a helper to resolve the endpoint.
            - For single-write accounts, it fails over between the first two write locations.
            - For other cases, it filters locations based on user and circuit-breaker exclusions.
        3.  Preferred Locations (Default): It uses the pre-filtered list of applicable read or write locations,
            respecting preferred locations, user exclusions, and circuit-breaker status, and selects an endpoint
            based on the request's location index.

        :param request: The request object for the current operation.
        :type request: azure.cosmos._request_object.RequestObject
        :return: The resolved endpoint URL as a string.
        :rtype: str
        """

        if request.location_endpoint_to_route:
            return request.location_endpoint_to_route

        location_index = int(request.location_index_to_route) if request.location_index_to_route else 0
        use_preferred_locations = (
            request.use_preferred_locations if request.use_preferred_locations is not None else True
        )

        is_write = documents._OperationType.IsWriteOperation(request.operation_type)

        if not use_preferred_locations or (
                is_write and not self.can_use_multiple_write_locations_for_request(request)
        ):
            return self._resolve_endpoint_without_preferred_locations(request, is_write, location_index)

        regional_routing_contexts = (
            self._get_applicable_write_regional_routing_contexts(request)
            if is_write
            else self._get_applicable_read_regional_routing_contexts(request)
        )
        regional_routing_context = regional_routing_contexts[location_index % len(regional_routing_contexts)]
        return regional_routing_context.get_primary()

    def should_refresh_endpoints(self):  # pylint: disable=too-many-return-statements
        most_preferred_location = self.effective_preferred_locations[0] if self.effective_preferred_locations else None

        # we should schedule refresh in background if we are unable to target the user's most preferredLocation.
        if self.connection_policy.EnableEndpointDiscovery:

            should_refresh = (self.connection_policy.UseMultipleWriteLocations
                              and not self.enable_multiple_writable_locations)

            if (most_preferred_location and most_preferred_location in
                    self.account_read_regional_routing_contexts_by_location):
                if (self.account_read_regional_routing_contexts_by_location
                        and most_preferred_location in self.account_read_regional_routing_contexts_by_location):
                    most_preferred_read_endpoint = (
                        self.account_read_regional_routing_contexts_by_location)[most_preferred_location]
                    if (most_preferred_read_endpoint and
                            most_preferred_read_endpoint != self.read_regional_routing_contexts[0]):
                        # For reads, we can always refresh in background as we can alternate to
                        # other available read endpoints
                        return True
                else:
                    return True

            if not self.can_use_multiple_write_locations():
                if self.is_endpoint_unavailable(self.write_regional_routing_contexts[0].get_primary(),
                                                EndpointOperationType.WriteType):
                    # same logic as other
                    # Since most preferred write endpoint is unavailable, we can only refresh in background if
                    # we have an alternate write endpoint
                    return True
                return should_refresh
            if (most_preferred_location and
                    most_preferred_location in self.account_write_regional_routing_contexts_by_location):
                most_preferred_write_regional_endpoint = (
                    self.account_write_regional_routing_contexts_by_location)[most_preferred_location]
                if most_preferred_write_regional_endpoint:
                    should_refresh |= most_preferred_write_regional_endpoint != self.write_regional_routing_contexts[0]
                    return should_refresh
                return True
            return should_refresh
        return False

    def is_endpoint_unavailable(self, endpoint: str, expected_available_operation: str):
        unavailability_info = (
            self.location_unavailability_info_by_endpoint[endpoint]
            if endpoint in self.location_unavailability_info_by_endpoint
            else None
        )

        if (
            expected_available_operation == EndpointOperationType.NoneType
            or not unavailability_info
            or expected_available_operation not in unavailability_info["operationType"]
        ):
            return False

        # Endpoint is unavailable
        return True

    def mark_endpoint_unavailable(
            self, unavailable_endpoint: str, unavailable_operation_type: EndpointOperationType, refresh_cache: bool,
            context: str):
        logger.warning("Marking %s unavailable for %s. Source: %s",
                       unavailable_endpoint,
                       unavailable_operation_type,
                       context)
        unavailability_info = (
            self.location_unavailability_info_by_endpoint[unavailable_endpoint]
            if unavailable_endpoint in self.location_unavailability_info_by_endpoint
            else None
        )
        if not unavailability_info:
            self.location_unavailability_info_by_endpoint[unavailable_endpoint] = {
                "operationType": set([unavailable_operation_type])
            }
        else:
            unavailable_operations = set([unavailable_operation_type]).union(unavailability_info["operationType"])
            self.location_unavailability_info_by_endpoint[unavailable_endpoint] = {
                "operationType": unavailable_operations
            }

        if refresh_cache:
            self.update_location_cache()

    def mark_endpoint_available(self, available_endpoint: str):
        self.location_unavailability_info_by_endpoint.pop(available_endpoint, "")

    def update_location_cache(self, write_locations=None, read_locations=None, enable_multiple_writable_locations=None):
        if enable_multiple_writable_locations:
            self.enable_multiple_writable_locations = enable_multiple_writable_locations

        if self.connection_policy.EnableEndpointDiscovery:
            if read_locations:
                (self.account_read_regional_routing_contexts_by_location,
                 self.account_locations_by_read_endpoints,
                 self.account_read_locations) = get_regional_routing_contexts_by_loc(read_locations)

            if write_locations:
                (self.account_write_regional_routing_contexts_by_location,
                 self.account_locations_by_write_endpoints,
                 self.account_write_locations) = get_regional_routing_contexts_by_loc(write_locations)

        # if preferred locations is empty and the default endpoint is a global endpoint,
        # we should use the read locations from gateway as effective preferred locations
        if self.connection_policy.PreferredLocations:
            self.effective_preferred_locations = self.connection_policy.PreferredLocations
        elif self.is_default_endpoint_regional():
            self.effective_preferred_locations = []
        elif not self.effective_preferred_locations:
            self.effective_preferred_locations = self.account_read_locations

        self.write_regional_routing_contexts = self.get_preferred_regional_routing_contexts(
            self.account_write_regional_routing_contexts_by_location,
            self.account_write_locations,
            EndpointOperationType.WriteType,
            self.default_regional_routing_context
        )
        self.read_regional_routing_contexts = self.get_preferred_regional_routing_contexts(
            self.account_read_regional_routing_contexts_by_location,
            self.account_read_locations,
            EndpointOperationType.ReadType,
            self.write_regional_routing_contexts[0]
        )

    def get_preferred_regional_routing_contexts(
        self, endpoints_by_location, orderedLocations, expected_available_operation, fallback_endpoint
    ):
        regional_endpoints = []
        # if enableEndpointDiscovery is false, we always use the defaultEndpoint that
        # user passed in during documentClient init
        if self.connection_policy.EnableEndpointDiscovery and endpoints_by_location:  # pylint: disable=too-many-nested-blocks
            if (
                self.can_use_multiple_write_locations()
                or expected_available_operation == EndpointOperationType.ReadType
            ):
                unavailable_endpoints = []
                if self.effective_preferred_locations:
                    # When client can not use multiple write locations, preferred locations
                    # list should only be used determining read endpoints order. If client
                    # can use multiple write locations, preferred locations list should be
                    # used for determining both read and write endpoints order.
                    for location in self.effective_preferred_locations:
                        regional_endpoint = endpoints_by_location.get(location)
                        if regional_endpoint:
                            if self.is_endpoint_unavailable(regional_endpoint.get_primary(),
                                                            expected_available_operation):
                                unavailable_endpoints.append(regional_endpoint)
                            else:
                                regional_endpoints.append(regional_endpoint)

                # If all preferred locations are unavailable, honor the preferred list by trying them anyway.
                if not regional_endpoints and unavailable_endpoints:
                    regional_endpoints.extend(unavailable_endpoints)

                # If there are no preferred locations or none of the preferred locations are in the account,
                # add the fallback endpoint.
                if not regional_endpoints:
                    regional_endpoints.append(fallback_endpoint)
            else:
                for location in orderedLocations:
                    if location and location in endpoints_by_location:
                        # location is empty during manual failover
                        regional_endpoint = endpoints_by_location[location]
                        regional_endpoints.append(regional_endpoint)

        if not regional_endpoints:
            regional_endpoints.append(fallback_endpoint)

        return regional_endpoints

    # if the endpoint is returned from the gateway in the account topology, it is a regional endpoint
    def is_default_endpoint_regional(self) -> bool:
        return any(
            context.get_primary() == self.default_regional_routing_context.get_primary()
            for context in self.account_read_regional_routing_contexts_by_location.values()
        )

    def can_use_multiple_write_locations(self):
        return self.connection_policy.UseMultipleWriteLocations and self.enable_multiple_writable_locations

    def can_use_multiple_write_locations_for_request(self, request):  # pylint: disable=name-too-long
        return self.can_use_multiple_write_locations() and (
            request.resource_type == ResourceType.Document
            or request.resource_type == ResourceType.PartitionKey
            or (
                request.resource_type == ResourceType.StoredProcedure
                and request.operation_type == documents._OperationType.ExecuteJavaScript
            )
        )

    def endpoints_to_health_check(self) -> Set[str]:
        # add read endpoints from gateway and in preferred locations
        health_check_endpoints = _get_health_check_endpoints(self.read_regional_routing_contexts)
        # add first write endpoint in case that the write region is not in preferred locations
        health_check_endpoints = health_check_endpoints.union(
            _get_health_check_endpoints(self.write_regional_routing_contexts[:1]
        ))

        return health_check_endpoints

    # get at most two regional routing contexts in the order from gateway and the ones specified in preferred locations
    @staticmethod
    def GetLocationalEndpoint(default_endpoint, location_name):
        # For default_endpoint like 'https://contoso.documents.azure.com:443/' parse it to
        # generate URL format. This default_endpoint should be global endpoint(and cannot
        # be a locational endpoint) and we agreed to document that
        endpoint_url = urlparse(default_endpoint)

        # hostname attribute in endpoint_url will return 'contoso.documents.azure.com'
        if endpoint_url.hostname is not None:
            hostname_parts = str(endpoint_url.hostname).lower().split(".")
            if hostname_parts is not None:
                # global_database_account_name will return 'contoso'
                global_database_account_name = hostname_parts[0]

                # Prepare the locational_database_account_name as contoso-eastus for location_name 'east us'
                locational_database_account_name = global_database_account_name + "-" + location_name.replace(" ", "")
                locational_database_account_name = locational_database_account_name.lower()

                # Replace 'contoso' with 'contoso-eastus' and return locational_endpoint
                # as https://contoso-eastus.documents.azure.com:443/
                locational_endpoint = default_endpoint.lower().replace(
                    global_database_account_name, locational_database_account_name, 1
                )
                return locational_endpoint

        return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey

import uuid
import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic throughput bucket operations at the client, database, container and item levels.
#
# 1. Setting throughput buckets at the Client Level
#
# 2. Setting Throughput Buckets at the Item Level
#    2.1 - Read Item
#    2.2 - Create Item
#
# 3. Multi-Bucket Usage
#    3.1 - Create and Delete Item with Separate Buckets
#    3.2 - Create Client and Create Item with Separate Buckets
#    3.3 - Create, Upsert, and Delete Item with Separate Buckets
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Databases and Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']

# Applies throughput bucket 1 to all requests from a client application
async def create_client_with_throughput_bucket(host=HOST, master_key=MASTER_KEY):
    async with CosmosClient(host, master_key, throughput_bucket=1) as client:
        pass

# Applies throughput bucket 2 for read item requests
async def container_read_item_throughput_bucket(client):
    database = client.get_database_client(DATABASE_ID)
    created_container = await database.create_container(
        str(uuid.uuid4()),
        PartitionKey(path="/pk"))
    created_document = await created_container.create_item(body={'id': '1' + str(uuid.uuid4()), 'pk': 'mypk'})

    await created_container.read_item(
         item=created_document['id'],
         partition_key="mypk",
         throughput_bucket=2)

    await database.delete_container(created_container.id)

# Applies throughput bucket 3 for create item requests
async def container_create_item_throughput_bucket(client):
    database = client.get_database_client(DATABASE_ID)

    created_container = await database.create_container(
        str(uuid.uuid4()),
        PartitionKey(path="/pk"))

    await created_container.create_item(
        body={'id': '1' + str(uuid.uuid4()), 'pk': 'mypk'},
        throughput_bucket=3)

    await database.delete_container(created_container.id)

# Applies throughput bucket 3 for create item requests and bucket 4 for delete item requests
async def container_create_and_delete_item_throughput_bucket(client):
    database = client.get_database_client(DATABASE_ID)

    created_container = await database.create_container(
        str(uuid.uuid4()),
        PartitionKey(path="/pk"))

    created_item = await created_container.create_item(
        body={'id': '1' + str(uuid.uuid4()), 'pk': 'mypk'},
        throughput_bucket=3)

    await created_container.delete_item(
        created_item['id'],
        partition_key='mypk',
        throughput_bucket=4)

    await database.delete_container(created_container.id)

# Applies throughput bucket 1 to all requests from a client application, and bucket 2 to create item requests
async def create_client_and_item_with_throughput_bucket(host=HOST, master_key=MASTER_KEY):
    async with CosmosClient(host, master_key,
        throughput_bucket=1) as client:

        database = client.get_database_client(DATABASE_ID)

        created_container = await database.create_container(
            str(uuid.uuid4()),
            PartitionKey(path="/pk"))

        await created_container.create_item(
            body={'id': '1' + str(uuid.uuid4()), 'pk': 'mypk'},
            throughput_bucket=2)

        await database.delete_container(created_container.id)

# Applies throughput bucket 3 for create item requests, bucket 4 for upsert item requests, and bucket 5 for delete item
async def container_create_upsert_and_delete_item_throughput_bucket(client):
    database = client.get_database_client(DATABASE_ID)

    created_container = await database.create_container(
        str(uuid.uuid4()),
        PartitionKey(path="/pk"))

    created_item = await created_container.create_item(
        body={'id': '1' + str(uuid.uuid4()), 'pk': 'mypk'},
        throughput_bucket=3)

    # add items for partition key 1
    for i in range(1, 3):
        await created_container.upsert_item(
            dict(id="item{}".format(i), pk='mypk', throughput_bucket=4))

    await created_container.delete_item(
        created_item['id'],
        partition_key='mypk',
        throughput_bucket=5)
    await database.delete_container(created_container.id)

async def run_sample():
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY} ) as client:
        await client.create_database_if_not_exists(id=DATABASE_ID)
        try:
            # creates client
            await create_client_with_throughput_bucket()

            # reads an item from a container
            await container_read_item_throughput_bucket(client)

            # writes an item to a container
            await container_create_item_throughput_bucket(client)

            # creates and deletes an item to a container
            await container_create_and_delete_item_throughput_bucket(client)

            # creates a client and item with separate throughput buckets
            await create_client_and_item_with_throughput_bucket()

            # creates an item, upserts multiple items, and deletes an item all on separate throughput buckets
            await container_create_upsert_and_delete_item_throughput_bucket(client)

        except exceptions.CosmosHttpResponseError as e:
            print('\nrun_sample has caught an error. {0}'.format(e.message))

        finally:
            print("\nrun_sample done")

if __name__ == '__main__':
    asyncio.run(run_sample())



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

"""Internal class for timeout failover retry policy implementation in the Azure
Cosmos database service.
"""
from azure.cosmos.documents import _OperationType


class _TimeoutFailoverRetryPolicy(object):

    def __init__(self, connection_policy, global_endpoint_manager, pk_range_wrapper, *args):
        self.retry_after_in_milliseconds = 500
        self.args = args
        self.request = args[0] if args else None

        self.global_endpoint_manager = global_endpoint_manager
        self.pk_range_wrapper = pk_range_wrapper
        # If an account only has 1 region, then we still want to retry once on the same region
        # We want this to be the default retry attempts as paging through a query means there are requests without
        # a request object
        self._max_retry_attempt_count = len(self.global_endpoint_manager.location_cache
                                            .read_regional_routing_contexts) + 1
       # If the request is a write operation, we only want to retry as many times as retry_write
        if self.request and _OperationType.IsWriteOperation(self.request.operation_type):
            self._max_retry_attempt_count = self.request.retry_write
        self.retry_count = 0
        self.connection_policy = connection_policy

    def ShouldRetry(self, _exception):
        """Returns true if the request should retry based on the passed-in exception.

        :param exceptions.CosmosHttpResponseError _exception:
        :returns: a boolean stating whether the request should be retried
        :rtype: bool
        """
        # we retry only if the request is a read operation or if it is a write operation with retry enabled
        if self.request and not self.is_operation_retryable():
            return False

        if not self.connection_policy.EnableEndpointDiscovery:
            return False

        self.retry_count += 1
        # Check if the next retry about to be done is safe
        if self.retry_count >= self._max_retry_attempt_count:
            return False

        # second check here ensures we only do cross-regional retries for read requests
        # non-idempotent write retries should only be retried once, using preferred locations if available (MM)
        if self.request and (_OperationType.IsReadOnlyOperation(self.request.operation_type)
                             or self.global_endpoint_manager.can_use_multiple_write_locations(self.request)):
            location_endpoint = self.resolve_next_region_service_endpoint()
            self.request.route_to_location(location_endpoint)
        return True

    # This function prepares the request to go to the next region
    def resolve_next_region_service_endpoint(self):
        # clear previous location-based routing directive
        self.request.clear_route_to_location()
        # set location-based routing directive based on retry count
        # ensuring usePreferredLocations is set to True for retry
        self.request.route_to_location_with_preferred_location_flag(self.retry_count, True)
        # Resolve the endpoint for the request and pin the resolution to the resolved endpoint
        # This enables marking the endpoint unavailability on endpoint failover/unreachability
        return self.global_endpoint_manager.resolve_service_endpoint_for_partition(self.request, self.pk_range_wrapper)

    def is_operation_retryable(self):
        if _OperationType.IsReadOnlyOperation(self.request.operation_type):
            return True
        return self.request.retry_write > 0



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.http_constants import StatusCodes
from azure.cosmos.partition_key import PartitionKey
import datetime

import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://learn.microsoft.com/azure/cosmos-db/nosql/quickstart-portal#create-account
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates Transactional Batch for Azure Cosmos DB Python SDK async
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = "batch_container"


async def execute_item_batch(database, container):
    print('\n1.11 Executing Batch Item operations\n')

    # We create three items to use for the sample. These are not part of the batch operations
    await container.create_item(get_sales_order("read_item"))
    await container.create_item(get_sales_order("delete_item"))
    await container.create_item(get_sales_order("replace_item"))
 
    # We create our batch operations
    create_item_operation = ("create", (get_sales_order("create_item"),))
    upsert_item_operation = ("upsert", (get_sales_order("upsert_item"),))
    read_item_operation = ("read", ("read_item",))
    delete_item_operation = ("delete", ("delete_item",))
    replace_item_operation = ("replace", ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                           "message": "item was replaced"}))
    replace_item_if_match_operation = ("replace",
                                       ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                         "message": "item was replaced"}),
                                       {"if_match_etag": container.client_connection.last_response_headers.get("etag")})
    replace_item_if_none_match_operation = ("replace",
                                            ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                              "message": "item was replaced"}),
                                            {"if_none_match_etag":
                                                 container.client_connection.last_response_headers.get("etag")})

    # Put our operations into a list
    batch_operations = [
        create_item_operation,
        upsert_item_operation,
        read_item_operation,
        delete_item_operation,
        replace_item_operation,
        # This below operation fails with status code 412, causing batch to fail and all operations to roll back
        replace_item_if_match_operation, # -> Comment this line out to see batch operations succeeding.
        replace_item_if_none_match_operation]

    # Run that list of operations
    try:
        # Batch results are returned as a list of item operation results - or raise a CosmosBatchOperationError if
        # one of the operations failed within your batch request.
        batch_results = await container.execute_item_batch(batch_operations=batch_operations, partition_key="Account1")
        print("\nResults for the batch operations: {}\n".format(batch_results))

    # For error handling, use try/ except with CosmosBatchOperationError and use the information in the
    # error returned for your application debugging, making it easy to pinpoint the failing operation
    except exceptions.CosmosBatchOperationError as e:
        error_operation_index = e.error_index
        error_operation_response = e.operation_responses[error_operation_index]
        error_operation = batch_operations[error_operation_index]
        print("\nError operation: {}, error operation response: {}\n".format(error_operation, error_operation_response))
        print("\nAn error occurred in the batch operation. All operations have been rolled back.\n")

    
    # You can also use this logic to read directly from a file into the batch you'd like to create:
    # with open("file_name.txt", "r") as data_file:
    #    container.execute_item_batch([("upsert", (t,)) for t in data_file.readlines()])


def get_sales_order(item_id):
    order1 = {'id': item_id,
              'account_number': 'Account1',
              'purchase_order_number': 'PO18009186470',
              'order_date': datetime.date(2005, 1, 10).strftime('%c'),
              'subtotal': 419.4589,
              'tax_amount': 12.5838,
              'freight': 472.3108,
              'total_due': 985.018,
              'items': [
                  {'order_qty': 1,
                   'product_id': 100,
                   'unit_price': 418.4589,
                   'line_price': 418.4589
                   }
              ],
              'ttl': 60 * 60 * 24 * 30
              }

    return order1


async def run_sample():
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        try:
            # setup database for this sample
            db = await client.create_database_if_not_exists(id=DATABASE_ID)

            # setup container for this sample
            container = await db.create_container_if_not_exists(id="batch_container",
                                                              partition_key=PartitionKey(path='/account_number'))
            print('Container with id \'{0}\' created'.format(CONTAINER_ID))

            await execute_item_batch(db, container)

            # cleanup database after sample
            try:
                await client.delete_database(db)

            except exceptions.CosmosResourceNotFoundError:
                pass

        except exceptions.CosmosHttpResponseError as e:
            print('\nrun_sample has caught an error. {0}'.format(e.message))

        finally:
            print("\nrun_sample done")


if __name__ == '__main__':
    asyncio.run(run_sample())
